
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdio.h>                          /* printf() */
#include <stdlib.h>                         /* malloc() free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <assert.h>

#include "initng_main.h"
#include "active_db.h"
#include "initng_load_module.h"
#include "initng_plugin_callers.h"
#include "initng_common.h"
#include "initng_toolbox.h"
#include "string_tools.h"

static int service_blacklisted(const char *name);

/* return TRUE if service is okay to load */
static int service_blacklisted(const char *name)
{
    int i;

    assert(name);
    /* walk through arguments looking for this dep to be blacklisted */
    for (i = 0; (g.Argv)[i]; ++i)
    {
        /* if we got a match */
        if ((g.Argv)[i][0] == '-' && !strcmp((g.Argv)[i] + 1, name))
        {
            return (TRUE);
        }
    }
    return (FALSE);
}

/* creates a active service */
int load_to_active(active_h ** a_new, const char *service_name)
{
    assert(service_name);
    D_("load_to_active(%s); \n", service_name);

    /* check the blacklist db */
    if (service_blacklisted(service_name))
    {
        F_("load_to_active(%s): Service BLACKLISTED.\n", service_name);
        return (FALSE);
    }

    /* Make sure there isn't any process with this name running */
    if ((*a_new = active_db_find_by_name(service_name)))
    {
        D_("load_to_active(%s): process exists, returning pointer to it. \n",
           service_name);
        return (TRUE);
    }

    /* if not create a new active entry, out of memory? */
    if (!active_db_new(a_new, service_name))
    {
        D_("load_to_active(%s): Unable to allocate process, out of memory?\n",
           service_name);
        (*a_new) = NULL;
        return (FALSE);
    }

    /* this actually loads service from disk */
    if (!get_service(*a_new) || !(*a_new)->from_service)
    {
        F_("load_to_active(%s): Can't set service!\n", service_name);
        active_db_free(*a_new);
        (*a_new) = NULL;
        return (FALSE);
    }

    if ((*a_new)->from_service->type != TYPE_SERVICE &&
        (*a_new)->from_service->type != TYPE_RUNLEVEL)
    {
        F_("load_to_active(%s): service is not a service type!\n",
           service_name);
        active_db_free(*a_new);
        (*a_new) = NULL;
        return (FALSE);
    }


    /* return with pointer to service */
    if (active_db_add(a_new))
        return (TRUE);
    else
        return (FALSE);

}



/* this function tries to find the from_service pointer from a service name.
   this actually loads the config from disk, parser by parser */
int get_service(active_h * service)
{
    assert(service);
    assert(service->name);
    D_("get_service(%s);\n", service->name);

    /* check if service is set already */
    if (service->from_service)
    {
        D_("get_service(%s): already have a service!\n", service->name);
        return (TRUE);
    }

    /* check cache database */
    if ((service->from_service = service_db_find_by_name(service->name)))
    {
        D_("get_service(%s): found service in service database!\n",
           service->name);
        return (TRUE);
    }

    /* try parse with dynamic loaded parsers */
    if ((service->from_service = initng_modules_parse(service->name)))
    {
        D_(" get_service(%s): got service from initng dynamic parsers.\n",
           service->name);
        return (TRUE);
    }

    /* or fail */
    D_("get_service(%s): Can't get source.. \n", service->name);
    return (FALSE);
}
static void dep_failed_to_start(active_h * service)
{
    active_h *next = NULL;

    while ((next = active_db_next_dep(service, next)))
    {
        mark_service(next, START_DEP_FAILED);
        dep_failed_to_start(next);
    }
}

static void dep_failed_to_stop(active_h * service)
{
    active_h *next = NULL;

    while ((next = active_db_next_dep(service, next)))
    {
        mark_service(next, STOP_DEP_FAILED);
        dep_failed_to_start(next);
    }
}

/* will return false, if plugins won't set the requested status */
int mark_service(active_h * service, e_a_status status)
{
    time_t old_time;

    assert(service);
    /* if already set */
    if (service->a_status == status)
    {
        D_("warning, this status %i is already set on %s!\n", status,
           service->name);
        /* it is actually already set to requested status */
        return (TRUE);
    }


    D_("mark_service(%s, %s);\n", service->name,
       active_db_get_status_string(status));

    /* set status and time */
    service->a_status = status;
    old_time = service->time_got_status;
    service->time_got_status = time(0);

    /* call all dynamic modules that want to know about this */
    initng_load_module_service_changed(service);

    /* is it really changed? */
    if (service->a_status != status)
    {
        D_("Some plugin did not want this to change!\n");
        service->time_got_status = old_time;
        return (FALSE);
    }

    /* add to history */
    history_add_values(service->name, service->time_got_status, status);


    /* if g.interrupt is set, main loop will try
       start new services */
    switch (service->a_status)
    {
        case DONE:
        case RUNNING:
            /* check if this is the final service */
            if (strcmp(service->name, g.runlevel) == 0)
                set_sys_state(STATE_UP);
            /* set interrupt, so main loop will check other
               that waits for deps */
            g.interrupt = TRUE;
            break;
        case STOPPED:
            /* set interrupt, so main loop will check other
               that waits for deps */
            g.interrupt = TRUE;
            break;
        case FAIL_STARTING:
            dep_failed_to_start(service);
            g.interrupt = TRUE;
            break;
        case FAIL_STOPPING:
            dep_failed_to_stop(service);
            g.interrupt = TRUE;
            break;
        case STOP_DEP_FAILED:
        case START_DEP_FAILED:
            g.interrupt = TRUE;
            break;
        default:
            break;
    }

    return (TRUE);
}



char *fix_variables(char *from, active_h * s)
{
    int i = 0;
    int j = 0;
    char *tmp;
    char *to;

    if (!from)
        return (NULL);
    assert(s);
    assert(s->name);
    to = (char *) initng_calloc((strlen(from) + 150), sizeof(char));
    while (from[i] != '\0')
    {
        if (from[i] == '$')
        {
            i++;
            if (strncmp(&from[i], "NAME", strlen("NAME")) == 0)
            {
                tmp = st_strip_path(s->name);
                strcpy(&to[j], tmp);
                i = i + strlen("NAME");
                j = j + strlen(tmp);
            }
            else if (strncmp(&from[i], "SERVICE", strlen("SERVICE")) == 0)
            {
                strcpy(&to[j], s->name);
                i = i + strlen("SERVICE");
                j = j + strlen(s->name);
            }
            else if (strncmp(&from[i], "CLASS", strlen("CLASS")) == 0)
            {
                tmp = st_get_path(s->name);
                strcpy(&to[j], tmp);
                i = i + strlen("CLASS");
                j = j + strlen(tmp);
                if (tmp)
                    free(tmp);
            }
            else
            {
                to[j] = from[i];
                j++;
                i++;
            }

        }
        else
        {
            to[j] = from[i];
            j++;
            i++;
        }
    }
    D_("fix_variables(): from \"%s\" to \"%s\".\n", from, to);
    return (to);
}
