#include <stdlib.h>
#include <sys/stat.h>
#include <stdio.h>
#include <dlfcn.h>
#include <assert.h>
#include <string.h>
#include <dirent.h>
#include <fnmatch.h>
#include <errno.h>
#include "initng.h"
#include "service_db.h"
#include "initng_load_module.h"
#include "initng_module.h"
#include "initng_toolbox.h"
#include "initng_global.h"
#include "initng_main.h"

/* Static includes */
#ifdef STATIC_CHROOT
#include "../plugins/chroot/initng_chroot.h"
#endif

/* this will null all arrays */
void initng_load_module_init(void)
{
    int i;

    for (i = 0; i < MAX_MODULES; i++)
    {
        g.modules[i].module_name = NULL;
        g.modules[i].module = NULL;
    }

    for (i = 0; i < MAX_CALLERS; i++)
        g.callers[i].ptype = NOPLUG;
}

m_h *initng_get_next_module(m_h * c)
{
    int i = 0;

    /* fast forward */
    if (c)
        for (; i < MAX_MODULES; i++)
            if (&g.modules[i] == c)
            {
                i++;
                break;
            }

    for (; i < MAX_MODULES; i++)
        if (g.modules[i].module_name)
            return (&g.modules[i]);

    return (NULL);
}

/* load a dynamic module */
int initng_load_module(const char *module_name)
{
    int i;
    int ret = 0;
    struct stat st;

    assert(module_name);

    /* check that file exists */
    if (stat(module_name, &st) != 0 || S_ISREG(st.st_mode) != 1)
    {
        F_("File \"%s\" not found, or not a regular file!\n\n", module_name);
        return (FALSE);
    }

    /* Try to find duplicates */
    for (i = 0; i < MAX_MODULES; i++)
        if (g.modules[i].module_name
            && strcmp(g.modules[i].module_name, module_name) == 0)
        {
            F_("Module \"%s\" already loaded, won't load it twice!\n\n",
               module_name);
            return (FALSE);
        }


    /* find a free entry */
    for (i = 0; i < MAX_MODULES; i++)
        if (!g.modules[i].module)
            break;

    /* Clear existing errors */
    dlerror();
    /* open module */
    g.modules[i].module = dlopen(module_name, RTLD_LAZY);
    /* this breaks ngc2 on my testbox - neuron : g.modules[i].module = dlopen(module_name, RTLD_NOW | RTLD_GLOBAL); */


    if (!g.modules[i].module)
    {
        F_("%s\n\n", dlerror());
        return (FALSE);
    }

    D_("Succeeded to open lib \"%s\"\n", module_name);

    /* Clear existing errors */
    dlerror();

    /* Load modules */
    g.modules[i].module_init = dlsym(g.modules[i].module, "module_init");
    if (!g.modules[i].module_init)
    {
        F_("%s\n\n", dlerror());
        return (FALSE);
    }

    g.modules[i].module_unload = dlsym(g.modules[i].module, "module_unload");
    if (!g.modules[i].module_unload)
    {
        F_("%s\n\n", dlerror());
        return (FALSE);
    }

    /* set module name in database */
    g.modules[i].module_name = i_strdup(module_name);

    if (g.modules[i].module_init)
        ret = (*g.modules[i].module_init) ();

    D_("for module \"%s\" return: %i\n", g.modules[i].module_name, ret);

    if (ret < 1)
    {
        F_("module %s did not load correctly (ret code %i), unloading in 1 sec...\n", g.modules[i].module_name, ret);
        sleep(1);
        initng_unload_module(&g.modules[i]);
    }

    return (TRUE);
}

int initng_unload_module_named(const char *name)
{
    int i;

    D_("initng_load_module_named(%s);\n", name);

    for (i = 0; i < MAX_MODULES; i++)
        if (g.modules[i].module && g.modules[i].module_name
            && strcmp(g.modules[i].module_name, name) == 0)
        {
            return (initng_unload_module(&g.modules[i]));
        }
    return (FALSE);
}

int initng_unload_module(m_h * module)
{
    assert(module);

    if (module->module_unload)
        (*module->module_unload) ();

    if (module->module != initng_load_static_modules)
        dlclose(module->module);
    module->module = NULL;
    if (module->module_name)
        free(module->module_name);
    module->module_name = NULL;
    return (TRUE);
}

void initng_load_all_modules(void)
{
    DIR *d;
    struct dirent *e;
    struct stat s;
    char *t = (char *) initng_calloc(100, sizeof(char));        /* allocate space for a filename */

    /* open plugin dir */
    if (!(d = opendir(INITNG_PLUGIN_DIR)))
    {
        F_("Unable top open plugin direcory " INITNG_PLUGIN_DIR ".\n");
        free(t);
        su_login();
    }

    /* get every entry */
    while ((e = readdir(d)) != NULL)
    {
        /* check for files, ending with .so.0.0.0 */
        if (fnmatch("*.so.0.0.0", e->d_name, 0) == 0)
        {
            /* set full path */
            strcpy(t, INITNG_PLUGIN_DIR "/");
            strcat(t, e->d_name);
            t[99] = '\0';                   /* make sure there is no buffer overrun */
            /* stat the file, to check that it exists, and is an ordinary file */
            if (stat(t, &s) == 0 && S_ISREG(s.st_mode))
            {
                D_("loading \"%s\"\n", t);
                initng_load_module(t);      /* load the module */
            }
        }
        else
        {
#ifdef DEBUG
            if (e->d_name[0] != '.')
            {
                D_("Won't load module \"%s\", doesn't match \"*.so.0.0.0\" pattern.\n", e->d_name);
            }
#endif
        }
    }
    closedir(d);
    free(t);


    initng_load_static_modules();
}

void initng_load_static_modules(void)
{
    int i;

    /* find a free entry */
    for (i = 0; i < MAX_MODULES; i++)
        if (!g.modules[i].module)
            break;

#ifdef STATIC_CHROOT
    W_("Warning, temporarily setting module_name to libchroot.so.0.0, this should be handled by a struct with name/version that checks name instead of module_name\n");
    strcpy(g.modules[i].module_name, "/lib/initng/libchroot.so.0.0");
    g.modules[i].module = initng_load_static_modules;
    g.modules[i].module_init = chroot_module_init;
    g.modules[i].module_unload = chroot_module_unload;
    i++;
#endif
}

/* this will dlclose all arrays */
void initng_unload_all_modules(void)
{
    int i;

    D_("initng_load_module_close_all(void);\n");

    for (i = 0; i < MAX_MODULES; i++)
        if (g.modules[i].module)
        {
            initng_unload_module(&g.modules[i]);
        }
}
